@@ -2,7 +2,7 @@
 
     Source file for  PGPsendmail  (wrapper to sendmail).
 
-    Copyright (C) 1994  Richard Gooch
+    Copyright (C) 1994,1995  Richard Gooch
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -66,7 +66,45 @@
     Updated by      Richard Gooch   30-JUN-1994: Made "GLOBAL" option the first
   set of flags to be processed.
 
-    Last updated by Richard Gooch   5-JUL-1994: Changed to use of  m_copy  .
+    Updated by      Richard Gooch   5-JUL-1994: Changed to use of  m_copy  .
+
+    Updated by      Richard Gooch   14-JUL-1994: Checked response flag that
+  PGPdaemon now returns.
+
+    Updated by      Richard Gooch   14-JUL-1994: Added more debug messages at
+  the end and discarded  /tmp/PGPsendmail.err.<PID>  files even if receipt is
+  discarded.
+
+    Updated by      Richard Gooch   20-JUL-1994: Added ability to sign all
+  outgoing messages (even those going out in plaintext).
+
+    Updated by      Richard Gooch   21-JUL-1994: Added declarations in
+  transmit_plaintext  for  errno  and  sys_errlist  .
+
+    Updated by      Richard Gooch   15-AUG-1994: Added  add_key  parameter to
+  transmit_plaintext  .
+
+    Updated by      Richard Gooch   22-AUG-1994: Automatically execute real
+  sendmail  if running as root (realUID = 0), and trap for invocation as
+  mailq  and  newaliases  .
+
+    Updated by      Richard Gooch  23-AUG-1994: Allow continuation of address
+  lines with trailing ','.
+
+    Updated by      Richard Gooch   21-NOV-1994: Supported  LOGNAME
+  environment variable.
+
+    Updated by      Richard Gooch   3-DEC-1994: Took account of changes to
+  connect_to_pgpdaemon  .
+
+    Updated by      Richard Gooch   6-DEC-1994: Wrote message in debug file
+  if no $PGPPATH/randseed.bin file exists.
+
+    Updated by      Richard Gooch   13-DEC-1994: Added support for address
+  format:  "Real Name"@email.address  in  extract_address  .
+
+    Last updated by Richard Gooch   17-DEC-1994: Fixed silly bug where username
+  was not set if no debug file existed.
 
 
 */
@@ -96,17 +134,22 @@ static int buf_len = 0;
 static char **recipients = NULL;
 static char *pgppath = NULL;
 static FILE *debug_fp = NULL;
+static char *username = NULL;
 
 
 STATIC_FUNCTION (flag read_line, (FILE *debug_fp) );
 STATIC_FUNCTION (flag process_addresses,
 		 (char *list, char *pgppath, option_type *options,
-		  FILE *debug_fp) );
+		  FILE *debug_fp, flag *trailing_comma) );
 STATIC_FUNCTION (void return_badnews, (char *message) );
 STATIC_FUNCTION (char *extract_address, (char *str, char **rest) );
 STATIC_FUNCTION (flag check_for_header_flags,
-		 (char *line, option_type *options, int sendmail_pid) );
+		 (char *line, option_type *options, int sendmail_pid,
+		  FILE *debug_fp) );
 STATIC_FUNCTION (void sig_handler, () );
+STATIC_FUNCTION (void transmit_plaintext, 
+		 (flag sign, FILE *debug_fp, int sendmail_pid,
+		  flag no_advertising, flag add_key) );
 
 
 void main (argc, argv, envp)
@@ -116,9 +159,12 @@ char *envp[];
 {
     int tmp;
 
+    flag bool_val;
     option_type options;
     flag header = TRUE;
     flag quoted_printables = FALSE;
+    flag continuation_line = FALSE;
+    flag dummy;
     int count;
     int ch;
     int tmpfd;
@@ -126,19 +172,25 @@ char *envp[];
     int out_fd, err_fd;
     int pgp_in_fd, pgp_pid;
     int status;
-    int pgpdaemon_fd;
+    int to_fd, from_fd;
     struct stat statbuf;
     char *bad_recipient;
-    FILE *out_fp;
     char *ptr1, *ptr2;
     char *real_sendmail = SENDMAIL;
     char txt[LINE_LENGTH];
     extern FILE *debug_fp;
+    extern char *buffer;
     extern char *pgppath;
+    extern char *username;
     extern char **recipients;
     ERRNO_TYPE errno;
     extern char *sys_errlist[];
 
+    if (argv[0] == NULL)
+    {
+	(void) fprintf (stderr, "argv[0] is NULL!\n");
+	exit (RV_UNDEF_ERROR);
+    }
     /*  Try to trap if SENDMAIL references PGPsendmail. I hope this works.
 	I only do this to protect people from themselves: if they *read* the
 	documentation, this wouldn't be necessary. Gosh, I'm such a nice
@@ -154,6 +206,12 @@ char *envp[];
 			"Someone didn't read the installation notes...\n");
 	exit (RV_UNDEF_ERROR);
     }
+    if ( (strcmp (ptr2, "mailq") == 0) || (strcmp (ptr2, "newaliases") == 0) )
+    {
+	(void) sleep (1);
+	(void) execve (real_sendmail, argv, envp);
+	exit (1);
+    }
     if ( ( pgppath = getenv ("PGPPATH") ) == NULL )
     {
 	/*  Delay 1 second in case the installer didn't read the instructions
@@ -166,12 +224,18 @@ char *envp[];
     if (getuid () == 0)
     {
 	/*  Hm. root shouldn't be using this.  */
+#ifdef DISALLOW_ROOT
+	(void) sleep (1);
+	(void) execve (real_sendmail, argv, envp);
+	exit (1);
+#else
 	(void) fprintf (stderr,
 			"It is not recommended for root to run PGPsendmail\n");
 	(void) fprintf (stderr, "with the environment variable PGPPATH set\n");
 	(void)fprintf(stderr,
 		      "Sleeping 15 seconds to give you a chance to kill me\n");
 	(void) sleep (15);
+#endif
     }
     (void) sprintf (txt, "%s/PGPsendmail.debug", pgppath);
     if (access (txt, F_OK) == 0)
@@ -201,8 +265,28 @@ char *envp[];
 	    (void) fprintf (debug_fp, "\"%s\"  ", argv[count]);
 	}
 	(void) fprintf (debug_fp, "\n");
+#ifdef NeXT
+	(void) fprintf (debug_fp, "Environment:\n");
+	for (count = 0; envp[count] != NULL; ++count)
+	{
+	    (void) fprintf (debug_fp, "  \"%s\"\n", envp[count]);
+	}
+#endif
 	(void) fflush (debug_fp);
     }
+    if ( ( username = getenv ("USER") ) == NULL )
+    {
+	if ( ( username = getenv ("LOGNAME") ) == NULL )
+	{
+	    if (debug_fp != NULL)
+	    {
+		(void) fprintf (debug_fp,
+				"No USER or LOGNAME environment variable!\n");
+		(void) fflush (debug_fp);
+	    }
+	    exit (RV_UNDEF_ERROR);
+	}
+    }
     (void) signal (SIGPIPE, sig_handler);
     options.secure = FALSE;
     options.insecure = FALSE;
@@ -220,8 +304,8 @@ char *envp[];
     if ( ( tmpfd = open (txt, O_CREAT | O_TRUNC | O_WRONLY, 0644) )
 	< 0 ) exit (1);
     if (dup2 (tmpfd, 2) == -1) exit (1);
-    (void) fprintf ( stderr, "To: %s\nSubject: PGPsendmail response\n\n",
-		    getenv ("USER") );
+    (void) fprintf (stderr, "To: %s\nSubject: PGPsendmail response\n\n",
+		    username);
     (void) fprintf (stderr, "PGPsendmail response message:\n\n");
     /*  Sanity checks for files the user should have  */
     (void) sprintf (txt, "%s/randseed.bin", pgppath);
@@ -229,6 +313,13 @@ char *envp[];
     {
 	(void) fprintf (stderr, "Could not find file: \"%s\"\t%s\n",
 			txt, sys_errlist[errno]);
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf(debug_fp,
+			   "Could not find file: \"%s\"\t%s\nPGP would hang\n",
+			   txt, sys_errlist[errno]);
+	    (void) fflush (debug_fp);
+	}
 	return_badnews ("PGP would hang");
 	exit (RV_SYS_ERROR);
     }
@@ -340,12 +431,8 @@ char *envp[];
 	return_badnews ("could not generate keylist");
 	exit (RV_SYS_ERROR);
     }
-    /*  Process header  */
-    if ( ( out_fp = fdopen (sendmail_in_fd, "w") ) == NULL )
-    {
-	return_badnews ("could not open output stream");
-	exit (1);
-    }
+    /*  Make stdout go to  sendmail  */
+    (void) dup2 (sendmail_in_fd, OUTPUT_FD);
     /*  An awful hack to insert fake recipients  */
     if (debug_fp != NULL)
     {
@@ -353,12 +440,13 @@ char *envp[];
 	(void) fflush (debug_fp);
     }
     (void) strcpy (txt, "pgp, -feat, +batchmode");
-    if ( !process_addresses (txt, NULL, NULL, debug_fp) )
+    if ( !process_addresses (txt, NULL, NULL, debug_fp, &dummy) )
     {
 	return_badnews ("add fake recipients");
 	kill (sendmail_pid, SIGKILL);
 	exit (1);
     }
+    /*  Process header  */
     if (debug_fp != NULL)
     {
 	(void) fprintf (debug_fp, "Starting to process header...\n");
@@ -382,17 +470,30 @@ char *envp[];
 			    buffer);
 	    (void) fflush (debug_fp);
 	}
-	if ( check_for_header_flags (buffer, &options,sendmail_pid) ) continue;
-	fputs (buffer, out_fp);
-	fputc ('\n', out_fp);
+	if ( check_for_header_flags (buffer, &options, sendmail_pid,
+				     debug_fp) ) continue;
+	fputs (buffer, stdout);
+	fputc ('\n', stdout);
 	if (buffer[0] == '\0')
 	{
 	    header = FALSE;
 	    continue;
 	}
+	if (continuation_line)
+	{
+	    if ( !process_addresses (buffer, pgppath, &options, debug_fp,
+				     &continuation_line) )
+	    {
+		return_badnews ("cannot process continuation header line");
+		kill (sendmail_pid, SIGKILL);
+		exit (1);
+	    }
+	    continue;
+	}
 	if (st_nicmp (buffer, "to:", 3) == 0)
 	{
-	    if ( !process_addresses (buffer + 3, pgppath, &options, debug_fp) )
+	    if ( !process_addresses (buffer + 3, pgppath, &options, debug_fp,
+				     &continuation_line) )
 	    {
 		return_badnews ("cannot process  to:  header line");
 		kill (sendmail_pid, SIGKILL);
@@ -402,7 +503,8 @@ char *envp[];
 	}
 	if (st_nicmp (buffer, "cc:", 3) == 0)
 	{
-	    if ( !process_addresses (buffer + 3, pgppath, &options, debug_fp) )
+	    if ( !process_addresses (buffer + 3, pgppath, &options, debug_fp,
+				     &continuation_line) )
 	    {
 		return_badnews ("cannot process  cc:  header line");
 		kill (sendmail_pid, SIGKILL);
@@ -412,7 +514,8 @@ char *envp[];
 	}
 	if (st_nicmp (buffer, "bcc:", 4) == 0)
 	{
-	    if ( !process_addresses (buffer + 4, pgppath, &options, debug_fp) )
+	    if ( !process_addresses (buffer + 4, pgppath, &options, debug_fp,
+				     &continuation_line) )
 	    {
 		return_badnews ("cannot process  bcc:  header line");
 		kill (sendmail_pid, SIGKILL);
@@ -430,7 +533,7 @@ char *envp[];
 	    (void) fprintf (stderr, "> %s\n\n", buffer);
 	}
     }
-    (void) fflush (out_fp);
+    (void) fflush (stdout);
     if (debug_fp != NULL)
     {
 	(void) fprintf (debug_fp, "Header processed\n");
@@ -444,12 +547,8 @@ char *envp[];
 			    "Insecure flag set: message will be sent as plaintext\n");
 	    (void) fflush (debug_fp);
 	}
-	/*  Copy message to  sendmail  and exit  */
-	while ( ( ch = fgetc (stdin) ) != EOF ) fputc (ch, out_fp);
-	fflush (out_fp);
-	(void) sprintf ( buffer, "/tmp/PGPsendmail.err.%d", getpid () );
-	(void) unlink (buffer);
-	exit (0);
+	transmit_plaintext (options.sign, debug_fp, sendmail_pid,
+			    options.no_advertising, options.add_key);
     }
     if (quoted_printables)
     {
@@ -466,12 +565,8 @@ char *envp[];
 			    "Quoted printables flag set: message will be sent as plaintext\n");
 	    (void) fflush (debug_fp);
 	}
-	/*  Copy message to  sendmail  and exit  */
-	while ( ( ch = fgetc (stdin) ) != EOF ) fputc (ch, out_fp);
-	fflush (out_fp);
-	(void) sprintf ( buffer, "/tmp/PGPsendmail.err.%d", getpid () );
-	(void) unlink (buffer);
-	exit (0);
+	transmit_plaintext (options.sign, debug_fp, sendmail_pid,
+			    options.no_advertising, options.add_key);
     }
     /*  Determine if all recipients have PGP public keys  */
     switch ( check_recipients (recipients + 3, pgppath, &bad_recipient) )
@@ -495,15 +590,9 @@ char *envp[];
 			    "Not all recipients have a public key: message will be sent as plaintext\n");
 	    (void) fflush (debug_fp);
 	}
-	/*  Copy message to  sendmail  and exit  */
-	while ( ( ch = fgetc (stdin) ) != EOF ) fputc (ch, out_fp);
-	fflush (out_fp);
-	(void) sprintf ( buffer, "/tmp/PGPsendmail.err.%d", getpid () );
-	(void) unlink (buffer);
-	exit (0);
-/*
+	transmit_plaintext (options.sign, debug_fp, sendmail_pid,
+			    options.no_advertising, options.add_key);
 	break;
-*/
       case CHECK_AMBIGUOUS_KEY:
 	if (options.secure)
 	{
@@ -521,15 +610,9 @@ char *envp[];
 			    "A recipient has an ambiguous public key: message will be sent as plaintext\n");
 	    (void) fflush (debug_fp);
 	}
-	/*  Copy message to  sendmail  and exit  */
-	while ( ( ch = fgetc (stdin) ) != EOF ) fputc (ch, out_fp);
-	fflush (out_fp);
-	(void) sprintf ( buffer, "/tmp/PGPsendmail.err.%d", getpid () );
-	(void) unlink (buffer);
-	exit (0);
-/*
+	transmit_plaintext (options.sign, debug_fp, sendmail_pid,
+			    options.no_advertising, options.add_key);
 	break;
-*/
       case CHECK_NO_KEYLIST:
 	return_badnews ("could not read keylist");
 	exit (RV_CANNOT_OPEN);
@@ -545,8 +628,6 @@ char *envp[];
 	(void) fprintf (debug_fp, "Message will be encrypted\n");
 	(void) fflush (debug_fp);
     }
-    /*  Make stdout go to  sendmail  */
-    dup2 (sendmail_in_fd, OUTPUT_FD);
     if (!options.no_advertising)
     {
 	(void) printf ("\nMessage has been automatically %sencrypted by PGPsendmail %s,\n",
@@ -558,19 +639,54 @@ char *envp[];
     if (options.sign)
     {
 	/*  Sign message  */
-	if ( ( pgpdaemon_fd = connect_to_pgpdaemon (pgppath) ) < 0 )
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "Will attempt to sign message\n");
+	    (void) fflush (debug_fp);
+	}
+	if ( !connect_to_pgpdaemon (pgppath, &to_fd, &from_fd) )
 	{
 	    return_badnews ("could not connect to PGPdaemon");
 	    kill (sendmail_pid, SIGKILL);
 	    exit (RV_UNDEF_ERROR);
 	}
-	if (write (pgpdaemon_fd, "SIGN\n", 5) < 5)
+	/*  Set close-on-exec flag for pipe to PGPdaemon  */
+	if (fcntl (to_fd, F_SETFD, 1) == -1)
+	{
+	    return_badnews ("could not set close-on-exec flag");
+	    kill (sendmail_pid, SIGKILL);
+	    exit (RV_UNDEF_ERROR);
+	}
+	if (write (to_fd, "SIGN\n", 5) < 5)
 	{
 	    return_badnews ("could not write to PGPdaemon");
 	    kill (sendmail_pid, SIGKILL);
 	    exit (RV_UNDEF_ERROR);
 	}
-	pgp_in_fd = pgpdaemon_fd;
+	/*  Check if OK so far  */
+	if (read (from_fd, (char *) &bool_val, sizeof bool_val) <
+	    sizeof bool_val)
+	{
+	    (void) fprintf (stderr,
+			    "Error reading response flag from PGPdaemon\t%s\n",
+			    sys_errlist[errno]);
+	    return_badnews ("problem communicating with PGPdaemon");
+	    kill (sendmail_pid, SIGKILL);
+	    exit (RV_UNDEF_ERROR);
+	}
+	if (!bool_val)
+	{
+	    return_badnews ("PGPdaemon refused to sign: probably has no passphrase");
+	    if (debug_fp != NULL)
+	    {
+		(void) fprintf (debug_fp,
+				"PGPdaemon refused to sign: probably has no passphrase\n");
+		(void) fflush (debug_fp);
+	    }
+	    kill (sendmail_pid, SIGKILL);
+	    exit (RV_UNDEF_ERROR);
+	}
+	pgp_in_fd = from_fd;
 	out_fd = OUTPUT_FD;
 	err_fd = ERROR_FD;
 	if ( ( pgp_pid = spawn_job ("pgp", recipients,
@@ -580,21 +696,16 @@ char *envp[];
 	    kill (sendmail_pid, SIGKILL);
 	    exit (RV_UNDEF_ERROR);
 	}
+	(void) close (from_fd);
 	/*  Copy data  */
-	if ( !copy_data (pgpdaemon_fd, INPUT_FD, FALSE) )
+	if ( !copy_data (to_fd, INPUT_FD, FALSE) )
 	{
 	    return_badnews ("could not copy data to be signed");
 	    kill (sendmail_pid, SIGKILL);
 	    exit (RV_UNDEF_ERROR);
 	}
 	/*  Tell PGPdaemon the end of the message has arrived  */
-	if (write (pgpdaemon_fd, "\0", 1) < 1)
-	{
-	    return_badnews ("could not copy data to be signed");
-	    kill (sendmail_pid, SIGKILL);
-	    exit (RV_UNDEF_ERROR);
-	}
-	(void) close (pgpdaemon_fd);
+	(void) close (to_fd);
 	(void) fprintf (stderr, "Signed message\n");
     }
     else
@@ -635,8 +746,32 @@ char *envp[];
 	}
     }
     (void) fprintf (stderr, "\n\n\n");
-    if (options.receipt) return_badnews (NULL);
-    exit (0);
+    if (debug_fp != NULL)
+    {
+	(void) fprintf (debug_fp,
+			"Message should have been sent to real  sendmail\n");
+	(void) fflush (debug_fp);
+    }
+    if (options.receipt)
+    {
+	return_badnews (NULL);
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "Receipt returned\n");
+	    (void) fflush (debug_fp);
+	}
+    }
+    else
+    {
+	(void) sprintf ( txt, "/tmp/PGPsendmail.err.%d", getpid () );
+	(void) unlink (txt);
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "Receipt discarded\n");
+	    (void) fflush (debug_fp);
+	}
+    }
+    exit (RV_OK);
 }   /*  End Function main  */
 
 static flag read_line (debug_fp)
@@ -697,7 +832,8 @@ FILE *debug_fp;
     return (FALSE);
 }   /*  End Function read_line  */
 
-static flag process_addresses (list, pgppath, options, debug_fp)
+static flag process_addresses (list, pgppath, options, debug_fp,
+			       trailing_comma)
 /*  This routine will process a line of addresses, adding each to the list of
     recipients.
     The line of addresses must be pointed to by  list  .
@@ -707,12 +843,15 @@ static flag process_addresses (list, pgppath, options, debug_fp)
     The storage pointed to by  options  may be modified.
     Debugging information will be written to the file pointed to by  debug_fp
     .If this is NULL, no debugging information is written.
+    If the line had a trailing comma, the value TRUE is written to the storage
+    pointed to by  trailing_comma  else the value FALSE is written here.
     The routine returns TRUE on succes, else it returns FALSE.
 */
 char *list;
 char *pgppath;
 option_type *options;
 FILE *debug_fp;
+flag *trailing_comma;
 {
     int new_size;
     char *ptr;
@@ -722,6 +861,11 @@ FILE *debug_fp;
     static int num_recipients = 0;
     static int recipient_buf_size = 0;
 
+    *trailing_comma = FALSE;
+    /*  Remove trailing whitespace and check for trailing ','  */
+    for (ptr = list + strlen (list) - 1; isspace (*ptr) && (ptr > list);
+	 --ptr) *ptr = '\0';
+    if (*ptr == ',') *trailing_comma = TRUE;
     while ( ( ptr = extract_address (list, &list) ) != NULL )
     {
 	if (pgppath == NULL)
@@ -794,15 +938,19 @@ char *message;
     char separator[SEPARATOR_LENGTH + 1];
     FILE *err_fp;
     char *argv[6];
+    extern char *username;
     ERRNO_TYPE errno;
     extern char *sys_errlist[];
 
+    /*  Make stdout go to  stderr  */
+    (void) fflush (stdout);
+    (void) dup2 (ERROR_FD, OUTPUT_FD);
     if (message == NULL)
     {
 	(void) fprintf (stderr, "PGPsendmail: PGP OK\n");
 	(void) sprintf ( txt,
-			"%s -t -f $USER -oi < /tmp/PGPsendmail.err.%d",
-			SENDMAIL, getpid () );
+			"%s -t -f %s -oi < /tmp/PGPsendmail.err.%d",
+			SENDMAIL, username, getpid () );
 	(void) system (txt);
 	(void) sprintf ( txt, "/tmp/PGPsendmail.err.%d", getpid () );
 	(void) unlink (txt);
@@ -820,7 +968,7 @@ char *message;
     argv[0] = SENDMAIL;
     argv[1] = "-t";
     argv[2] = "-f";
-    argv[3] = getenv ("USER");
+    argv[3] = username;
     argv[4] = "-oi";
     argv[5] = NULL;
     in_fd = -1;
@@ -877,6 +1025,7 @@ char **rest;
 	*rest = NULL;
 	return (NULL);
     }
+    *rest = NULL;
     /*  Skip leading whitespace  */
     while ( isspace (*str) && (*str != '\0' ) ) ++str;
     if (*str == ',') ++str;
@@ -884,15 +1033,37 @@ char **rest;
     while ( isspace (*str) && (*str != '\0' ) ) ++str;
     if (*str == '\"')
     {
+	/*  Address formats:
+	        "Real Name" <username@email.address>
+		"Real Name"@email.address
+	*/
+	if ( ( ptr1 = strchr (str + 1, '\"') ) == NULL )
+	{
+	    /*  Stuffed address  */
+	    (void) fprintf (stderr,
+			    "Bad address (missing close quote): \"%s\"\n",
+			    str);
+	    return_badnews ("bad address");
+	    exit (1);
+	}
+	if (ptr1[1] == '@')
+	{
+	    /*  Address format:
+		"Real Name"@email.address
+	    */
+	    if ( ( *rest = strchr (ptr1, ',') ) != NULL ) ++*rest;
+	    return (str);
+	}
 	/*  Address format:
 	        "Real Name" <username@email.address>
 	*/
 	/*  Discard bit in quotes  */
-	m_free ( ex_str (str, &str) );
+	str = ptr1 + 1;
 	if ( ( ptr1 = strchr (str, '<') ) == NULL )
 	{
 	    /*  Stuffed address  */
-	    (void) fprintf (stderr, "Bad address: \"%s\"\n", str);
+	    (void) fprintf (stderr, "Bad address (missing '<'): \"%s\"\n",
+			    str);
 	    return_badnews ("bad address");
 	    exit (1);
 	}
@@ -993,19 +1164,22 @@ char **rest;
     return (str);
 }   /*  End Function extract_address  */
 
-static flag check_for_header_flags (line, options, sendmail_pid)
+static flag check_for_header_flags (line, options, sendmail_pid, debug_fp)
 /*  This routine will check if the header line contains the "X-Secure:"
     keyword and will process it.
     The header line must be pointed to by  line  .
     The flags will modify the storage pointed to by  options  .See the
     documentation on the  process_flags  routine for details.
     The process ID of the real sendmail must be given by  sendmail_pid  .
+    Debugging information will be written to the file pointed to by  debug_fp
+    .If this is NULL, no debugging information is written.
     The routine returns TRUE if the header line contains the "X-Secure:"
     keyword, else it returns FALSE.
 */
 char *line;
 option_type *options;
 int sendmail_pid;
+FILE *debug_fp;
 {
     char *ptr, *rest;
 
@@ -1013,8 +1187,18 @@ int sendmail_pid;
     line += 9;
     if ( !process_flags (line, options) )
     {
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "bad X-Secure: line: \"%s\"\n", line);
+	    (void) fflush (debug_fp);
+	}
 	return_badnews ("bad X-Secure: line");
 	kill (sendmail_pid, SIGKILL);
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "Real  sendmail  killed: exiting\n");
+	    (void) fflush (debug_fp);
+	}
 	exit (1);
     }
     return (TRUE);
@@ -1043,10 +1227,217 @@ char *addr;
     }
 }   /*  End Function sig_handler  */
 
+static void transmit_plaintext (sign, debug_fp, sendmail_pid,
+				no_advertising, add_key)
+/*  This routine will transmit the Email message the message in plaintext form.
+    If the value of  sign  is TRUE, the message will be.
+    Debugging information will be written to the file pointed to by  debug_fp
+    .If this is NULL, no debugging information is written.
+    The process ID of the real sendmail must be given by  sendmail_pid  .
+    If advertising is not desired the value of  no_advertising  must be TRUE.
+    The routine does not return: upon completion the process exits.
+*/
+flag sign;
+FILE *debug_fp;
+int sendmail_pid;
+flag no_advertising;
+flag add_key;
+{
+    flag bool_val;
+    int ch;
+    int child_pid, status;
+    int to_fd, from_fd;
+    extern char *buffer;
+    extern char *pgppath;
+    ERRNO_TYPE errno;
+    extern char *sys_errlist[];
+
+    if (!sign)
+    {
+	if (debug_fp != NULL)
+	{
+	    (void) fputs ("Copying unsigned, plaintext message to  sendmail\n",
+			  debug_fp);
+	    (void) fflush (debug_fp);
+	}
+	/*  Copy message to  sendmail  */
+	while ( ( ch = fgetc (stdin) ) != EOF ) fputc (ch, stdout);
+	fflush (stdout);
+	if (add_key)
+	{
+	    if ( !include_mykey (pgppath, stdout) )
+	    {
+		return_badnews ("could not include public key");
+		kill (sendmail_pid, SIGKILL);
+		exit (1);
+	    }
+	}
+	/*  Clean up and exit  */
+	(void) sprintf ( buffer, "/tmp/PGPsendmail.err.%d", getpid () );
+	(void) unlink (buffer);
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "Exiting with status 0\n");
+	    (void) fflush (debug_fp);
+	}
+	exit (0);
+    }
+    /*  Message should be signed  */
+    if (debug_fp != NULL)
+    {
+	(void) fprintf (debug_fp, "Will attempt to sign message\n");
+	(void) fflush (debug_fp);
+    }
+    if ( !connect_to_pgpdaemon (pgppath, &to_fd, &from_fd) )
+    {
+	return_badnews ("could not connect to PGPdaemon");
+	kill (sendmail_pid, SIGKILL);
+	exit (RV_UNDEF_ERROR);
+    }
+    if (write (to_fd, "SIGN\n", 5) < 5)
+    {
+	return_badnews ("could not write to PGPdaemon");
+	kill (sendmail_pid, SIGKILL);
+	exit (RV_UNDEF_ERROR);
+    }
+    /*  Check if OK so far  */
+    if (read (from_fd, (char *) &bool_val, sizeof bool_val) < sizeof bool_val)
+    {
+	return_badnews ("problem communicating with PGPdaemon");
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp,
+			    "Error reading response flag from PGPdaemon\t%s\n",
+			    sys_errlist[errno]);
+	    (void) fflush (debug_fp);
+	}
+	kill (sendmail_pid, SIGKILL);
+	exit (RV_UNDEF_ERROR);
+    }
+    if (!bool_val)
+    {
+	return_badnews ("PGPdaemon refused to sign: probably has no passphrase");
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp,
+			    "PGPdaemon refused to sign: probably has no passphrase\n");
+	    (void) fflush (debug_fp);
+	}
+	kill (sendmail_pid, SIGKILL);
+	exit (RV_UNDEF_ERROR);
+    }
+    if (debug_fp != NULL)
+    {
+	(void) fprintf (debug_fp,
+			"Copying signed, plaintext message to  sendmail\n");
+	(void) fflush (debug_fp);
+    }
+    if (!no_advertising)
+    {
+	(void) printf ("\nMessage has been automatically signed by PGPsendmail %s,\n",
+		       VERSION);
+	(void) printf ("available from ftp.atnf.csiro.au:pub/people/rgooch");
+	(void) printf ("\n\n");
+	(void) fflush (stdout);
+    }
+    /*  Now have to copy stdin to PGPdaemon input and PGPdaemon output to
+	stdout  */
+    /*  Setup a child process. Parent copies stdin to PGPdaemon.
+	Child copies PGPdaemon output to stdout  */
+    switch ( child_pid = fork () )
+    {
+      case 0:
+	/*  Child  */
+	(void) close (to_fd);
+	if ( !copy_data (OUTPUT_FD, from_fd, FALSE) )
+	{
+	    (void) fprintf (stderr, "Error copying data from PGPdaemon\n");
+	    if (debug_fp != NULL)
+	    {
+		(void) fprintf (debug_fp,
+				"Error copying data from PGPdaemon\n");
+		(void) fflush (debug_fp);
+	    }
+	    exit (RV_UNDEF_ERROR);
+	}
+	/*  Done  */
+	(void) close (from_fd);
+	exit (RV_OK);
+/*
+	break;
+*/
+      case -1:
+	/*  Error  */
+	(void) fprintf (stderr, "Could not fork\t%s\n",
+			sys_errlist[errno]);
+	(void) copy_data (OUTPUT_FD, INPUT_FD, FALSE);
+	exit (RV_OK);
+/*
+	break;
+*/
+      default:
+	/*  Parent  */
+	(void) close (from_fd);
+	break;
+    }
+    /*  Copy  stdin  to PGPdaemon  */
+    if ( !copy_data (to_fd, INPUT_FD, FALSE) )
+    {
+	return_badnews ("Error copying data to PGPdaemon");
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "Error copying data to PGPdaemon\n");
+	    (void) fflush (debug_fp);
+	}
+	kill (sendmail_pid, SIGKILL);
+	exit (RV_UNDEF_ERROR);
+    }
+    /*  Tell PGPdaemon the end of message has arrived  */
+    (void) close (to_fd);
+    /*  Reap child  */
+    (void) waitpid (child_pid, &status, 0);
+    if (status != 0)
+    {
+	return_badnews ("Child returned bad status");
+	if (debug_fp != NULL)
+	{
+	    (void) fprintf (debug_fp, "Bad child status: %d\n", status);
+	    (void) fflush (debug_fp);
+	}
+	(void) kill (sendmail_pid, SIGKILL);
+	exit (RV_OK);
+    }
+    if (debug_fp != NULL)
+    {
+	(void) fprintf (debug_fp, "Signed message\n");
+	(void) fflush (debug_fp);
+    }
+    if (add_key)
+    {
+	if ( !include_mykey (pgppath, stdout) )
+	{
+	    return_badnews ("could not include public key");
+	    kill (sendmail_pid, SIGKILL);
+	    exit (1);
+	}
+    }
+    /*  Clean up and exit  */
+    (void) sprintf ( buffer, "/tmp/PGPsendmail.err.%d", getpid () );
+    (void) unlink (buffer);
+    if (debug_fp != NULL)
+    {
+	(void) fprintf (debug_fp, "Exiting with status 0\n");
+	(void) fflush (debug_fp);
+    }
+    exit (0);
+}   /*  End Function transmit_plaintext  */
+
+
 /*  Valid address formats:
 
     Real Name <username@email.address>
     "Real Name" <username@email.address>
+    "Real Name"@email.address
     username@email.address
     username
     username (Real Name [,Organisation])
